@@ -8,22 +8,25 @@
 
 #include <cmath>
 
-#include <QReadWriteLock>
 #include <QDir>
 #include <QFile>
 #include <QFileInfo>
+#include <QMimeDatabase>
+#include <QReadWriteLock>
 #include <QRegularExpression>
 #include <QSharedData>
-#include <QMimeDatabase>
 #include <QStandardPaths>
 #include <QUrl>
 
 #include <KConfigGroup>
 #include <KShell>
-#include <KProtocolInfo>
 
-#include "krunner_debug.h"
+#if KRUNNER_BUILD_DEPRECATED_SINCE(5, 76)
+#include <KProtocolInfo>
+#endif
+
 #include "abstractrunner.h"
+#include "krunner_debug.h"
 #include "querymatch.h"
 
 #define LOCK_FOR_READ(d) d->lock.lockForRead();
@@ -32,139 +35,160 @@
 
 namespace Plasma
 {
-
 class RunnerContextPrivate : public QSharedData
 {
-    public:
-        RunnerContextPrivate(RunnerContext *context)
-            : QSharedData(),
-              type(RunnerContext::UnknownType),
-              q(context),
-              singleRunnerQueryMode(false)
-        {
-        }
+public:
+    RunnerContextPrivate(RunnerContext *context)
+        : QSharedData()
+        , type(RunnerContext::UnknownType)
+        , q(context)
+    {
+    }
 
-        RunnerContextPrivate(const RunnerContextPrivate &p)
-            : QSharedData(),
-              launchCounts(p.launchCounts),
-              type(RunnerContext::None),
-              q(p.q),
-              singleRunnerQueryMode(false)
-        {
-        }
+    RunnerContextPrivate(const RunnerContextPrivate &p)
+        : QSharedData()
+        , launchCounts(p.launchCounts)
+        , type(RunnerContext::None)
+        , q(p.q)
+    {
+    }
 
-        ~RunnerContextPrivate()
-        {
-        }
+    ~RunnerContextPrivate()
+    {
+    }
 
 #if KRUNNER_BUILD_DEPRECATED_SINCE(5, 76)
-        /**
-         * Determines type of query
-                    &&
-         */
-        void determineType()
-        {
-            // NOTE! this method must NEVER be called from
-            // code that may be running in multiple threads
-            // with the same data.
-            type = RunnerContext::UnknownType;
-            QString path = QDir::cleanPath(KShell::tildeExpand(term));
+    /**
+     * Determines type of query
+                &&
+     */
+    void determineType()
+    {
+        // NOTE! this method must NEVER be called from
+        // code that may be running in multiple threads
+        // with the same data.
+        type = RunnerContext::UnknownType;
+        QString path = QDir::cleanPath(KShell::tildeExpand(term));
 
-            int space = path.indexOf(QLatin1Char(' '));
-            if (!QStandardPaths::findExecutable(path.left(space)).isEmpty()) {
-                // it's a shell command if there's a space because that implies
-                // that it has arguments!
-                type = (space > 0) ? RunnerContext::ShellCommand :
-                                     RunnerContext::Executable;
-            } else {
-                QUrl url = QUrl::fromUserInput(term);
+        int space = path.indexOf(QLatin1Char(' '));
+        if (!QStandardPaths::findExecutable(path.left(space)).isEmpty()) {
+            // it's a shell command if there's a space because that implies
+            // that it has arguments!
+            type = (space > 0) ? RunnerContext::ShellCommand : RunnerContext::Executable;
+        } else {
+            QUrl url = QUrl::fromUserInput(term);
 
-                // QUrl::fromUserInput assigns http to everything if it cannot match it to
-                // anything else. We do not want that.
-                if (url.scheme() == QLatin1String("http")) {
-                    if (!term.startsWith(QLatin1String("http"))) {
-                        url.setScheme(QString());
-                    }
+            // QUrl::fromUserInput assigns http to everything if it cannot match it to
+            // anything else. We do not want that.
+            if (url.scheme() == QLatin1String("http")) {
+                if (!term.startsWith(QLatin1String("http"))) {
+                    url.setScheme(QString());
                 }
+            }
 
-                const bool hasProtocol = !url.scheme().isEmpty();
-                const bool isLocalProtocol = !hasProtocol || KProtocolInfo::protocolClass(url.scheme()) == QLatin1String(":local");
-                if ((hasProtocol &&
-                    ((!isLocalProtocol && !url.host().isEmpty()) ||
-                     (isLocalProtocol && url.scheme() != QLatin1String("file"))))
-                        || term.startsWith(QLatin1String("\\\\"))) {
-                    // we either have a network protocol with a host, so we can show matches for it
-                    // or we have a non-file url that may be local so a host isn't required
-                    // or we have an UNC path (\\foo\bar)
-                    type = RunnerContext::NetworkLocation;
-                } else if (isLocalProtocol) {
-                    // at this point in the game, we assume we have a path,
-                    // but if a path doesn't have any slashes
-                    // it's too ambiguous to be sure we're in a filesystem context
-                    path = !url.scheme().isEmpty() ? QDir::cleanPath(url.toLocalFile()) : path;
-                    if ((path.indexOf(QLatin1Char('/')) != -1 || path.indexOf(QLatin1Char('\\')) != -1)) {
-                        QFileInfo info(path);
-                        if (info.isSymLink()) {
-                            path = info.canonicalFilePath();
-                            info = QFileInfo(path);
-                        }
-                        if (info.isDir()) {
-                            type = RunnerContext::Directory;
-                            mimeType = QStringLiteral("inode/folder");
-                        } else if (info.isFile()) {
-                            type = RunnerContext::File;
-                            QMimeDatabase db;
-                            QMimeType mime = db.mimeTypeForFile(path);
-                            if (!mime.isDefault()) {
-                                mimeType = mime.name();
-                            }
+            const bool hasProtocol = !url.scheme().isEmpty();
+            const bool isLocalProtocol = !hasProtocol || KProtocolInfo::protocolClass(url.scheme()) == QLatin1String(":local");
+            if ((hasProtocol && ((!isLocalProtocol && !url.host().isEmpty()) || (isLocalProtocol && url.scheme() != QLatin1String("file"))))
+                || term.startsWith(QLatin1String("\\\\"))) {
+                // we either have a network protocol with a host, so we can show matches for it
+                // or we have a non-file url that may be local so a host isn't required
+                // or we have an UNC path (\\foo\bar)
+                type = RunnerContext::NetworkLocation;
+            } else if (isLocalProtocol) {
+                // at this point in the game, we assume we have a path,
+                // but if a path doesn't have any slashes
+                // it's too ambiguous to be sure we're in a filesystem context
+                path = !url.scheme().isEmpty() ? QDir::cleanPath(url.toLocalFile()) : path;
+                if ((path.indexOf(QLatin1Char('/')) != -1 || path.indexOf(QLatin1Char('\\')) != -1)) {
+                    QFileInfo info(path);
+                    if (info.isSymLink()) {
+                        path = info.canonicalFilePath();
+                        info = QFileInfo(path);
+                    }
+                    if (info.isDir()) {
+                        type = RunnerContext::Directory;
+                        mimeType = QStringLiteral("inode/folder");
+                    } else if (info.isFile()) {
+                        type = RunnerContext::File;
+                        QMimeDatabase db;
+                        QMimeType mime = db.mimeTypeForFile(path);
+                        if (!mime.isDefault()) {
+                            mimeType = mime.name();
                         }
                     }
                 }
             }
         }
+    }
 #endif
 
-        void invalidate()
-        {
-            q = &s_dummyContext;
-        }
+    void invalidate()
+    {
+        q = &s_dummyContext;
+    }
 
-        QueryMatch* findMatchById(const QString &id)
-        {
-            QueryMatch *match = nullptr;
-            lock.lockForRead();
-            for (auto &matchEntry : matches) {
-                if (matchEntry.id() == id) {
-                    match = &matchEntry;
-                    break;
-                }
+#if KRUNNER_BUILD_DEPRECATED_SINCE(5, 81)
+    QueryMatch *findMatchById(const QString &id)
+    {
+        QueryMatch *match = nullptr;
+        lock.lockForRead();
+        for (auto &matchEntry : matches) {
+            if (matchEntry.id() == id) {
+                match = &matchEntry;
+                break;
             }
-            lock.unlock();
-            return match;
         }
+        lock.unlock();
+        return match;
+    }
+#endif
 
-        QReadWriteLock lock;
-        QList<QueryMatch> matches;
-        QHash<QString, int> launchCounts;
-        QString term;
-        QString mimeType;
-        QStringList enabledCategories;
-        RunnerContext::Type type;
-        RunnerContext * q;
-        static RunnerContext s_dummyContext;
-        bool singleRunnerQueryMode;
+    void addMatch(const QueryMatch &match)
+    {
+        if (match.runner() && match.runner()->hasUniqueResults()) {
+            if (uniqueIds.contains(match.id())) {
+                const QueryMatch &existentMatch = uniqueIds.value(match.id());
+                if (existentMatch.runner() && existentMatch.runner()->hasWeakResults()) {
+                    // There is an existing match with the same ID and we are allowed to replace it
+                    matches.removeOne(existentMatch);
+                    matches.append(match);
+                }
+            } else {
+                // There is no existing match with the same id
+                uniqueIds.insert(match.id(), match);
+                matches.append(match);
+            }
+        } else {
+            // Runner has the unique results property not set
+            matches.append(match);
+        }
+    }
+
+    QReadWriteLock lock;
+    QList<QueryMatch> matches;
+    QHash<QString, int> launchCounts;
+    QString term;
+    QString mimeType;
+    QStringList enabledCategories;
+    RunnerContext::Type type;
+    RunnerContext *q;
+    static RunnerContext s_dummyContext;
+    bool singleRunnerQueryMode = false;
+    bool shouldIgnoreCurrentMatchForHistory = false;
+    QMap<QString, QueryMatch> uniqueIds;
+    QString requestedText;
+    int requestedCursorPosition = 0;
 };
 
 RunnerContext RunnerContextPrivate::s_dummyContext;
 
 RunnerContext::RunnerContext(QObject *parent)
-    : QObject(parent),
-      d(new RunnerContextPrivate(this))
+    : QObject(parent)
+    , d(new RunnerContextPrivate(this))
 {
 }
 
-//copy ctor
+// copy ctor
 RunnerContext::RunnerContext(RunnerContext &other, QObject *parent)
     : QObject(parent)
 {
@@ -213,25 +237,28 @@ void RunnerContext::reset()
     // we still have to remove all the matches, since if the
     // ref count was 1 (e.g. only the RunnerContext is using
     // the dptr) then we won't get a copy made
-    if (!d->matches.isEmpty()) {
-        d->matches.clear();
-        Q_EMIT matchesChanged();
-    }
-
+    d->matches.clear();
     d->term.clear();
+    Q_EMIT matchesChanged();
+
     d->mimeType.clear();
+    d->uniqueIds.clear();
     d->type = UnknownType;
     d->singleRunnerQueryMode = false;
+    d->shouldIgnoreCurrentMatchForHistory = false;
 }
 
 void RunnerContext::setQuery(const QString &term)
 {
-    reset();
+    if (!this->query().isEmpty()) {
+        reset();
+    }
 
     if (term.isEmpty()) {
         return;
     }
 
+    d->requestedText.clear(); // Invalidate this field whenever the query changes
     d->term = term;
 #if KRUNNER_BUILD_DEPRECATED_SINCE(5, 76)
     d->determineType();
@@ -247,7 +274,7 @@ QString RunnerContext::query() const
 }
 #if KRUNNER_BUILD_DEPRECATED_SINCE(5, 76)
 
-void RunnerContext::setEnabledCategories(const QStringList& categories)
+void RunnerContext::setEnabledCategories(const QStringList &categories)
 {
     d->enabledCategories = categories;
 }
@@ -286,7 +313,7 @@ bool RunnerContext::isValid() const
 bool RunnerContext::addMatches(const QList<QueryMatch> &matches)
 {
     if (matches.isEmpty() || !isValid()) {
-       //Bail out if the query is empty or the qptr is dirty
+        // Bail out if the query is empty or the qptr is dirty
         return false;
     }
 
@@ -295,10 +322,9 @@ bool RunnerContext::addMatches(const QList<QueryMatch> &matches)
         // Give previously launched matches a slight boost in relevance
         // The boost smoothly saturates to 0.5;
         if (int count = d->launchCounts.value(match.id())) {
-            match.setRelevance(match.relevance() + 0.5 * (1-exp(-count*0.3)));
+            match.setRelevance(match.relevance() + 0.5 * (1 - exp(-count * 0.3)));
         }
-
-        d->matches.append(match);
+        d->addMatch(match);
     }
     UNLOCK(d);
     // A copied searchContext may share the d pointer,
@@ -323,14 +349,14 @@ bool RunnerContext::addMatch(const QueryMatch &match)
     if (int count = d->launchCounts.value(m.id())) {
         m.setRelevance(m.relevance() + 0.05 * count);
     }
-
-    d->matches.append(m);
+    d->addMatch(match);
     UNLOCK(d);
     Q_EMIT d->q->matchesChanged();
 
     return true;
 }
 
+#if KRUNNER_BUILD_DEPRECATED_SINCE(5, 81)
 bool RunnerContext::removeMatches(const QStringList matchIdList)
 {
     bool success = false;
@@ -342,14 +368,16 @@ bool RunnerContext::removeMatches(const QStringList matchIdList)
 
     return success;
 }
+#endif
 
+#if KRUNNER_BUILD_DEPRECATED_SINCE(5, 81)
 bool RunnerContext::removeMatch(const QString matchId)
 {
     if (!isValid()) {
         return false;
     }
     LOCK_FOR_READ(d)
-    const QueryMatch* match = d->findMatchById(matchId);
+    const QueryMatch *match = d->findMatchById(matchId);
     UNLOCK(d)
     if (!match) {
         return false;
@@ -361,7 +389,9 @@ bool RunnerContext::removeMatch(const QString matchId)
 
     return true;
 }
+#endif
 
+#if KRUNNER_BUILD_DEPRECATED_SINCE(5, 81)
 bool RunnerContext::removeMatches(Plasma::AbstractRunner *runner)
 {
     if (!isValid()) {
@@ -371,7 +401,7 @@ bool RunnerContext::removeMatches(Plasma::AbstractRunner *runner)
     QList<QueryMatch> presentMatchList;
 
     LOCK_FOR_READ(d)
-    for(const QueryMatch &match : qAsConst(d->matches)) {
+    for (const QueryMatch &match : std::as_const(d->matches)) {
         if (match.runner() == runner) {
             presentMatchList << match;
         }
@@ -383,7 +413,7 @@ bool RunnerContext::removeMatches(Plasma::AbstractRunner *runner)
     }
 
     LOCK_FOR_WRITE(d)
-    for (const QueryMatch &match : qAsConst(presentMatchList)) {
+    for (const QueryMatch &match : std::as_const(presentMatchList)) {
         d->matches.removeAll(match);
     }
     UNLOCK(d)
@@ -391,6 +421,7 @@ bool RunnerContext::removeMatches(Plasma::AbstractRunner *runner)
     Q_EMIT d->q->matchesChanged();
     return true;
 }
+#endif
 
 QList<QueryMatch> RunnerContext::matches() const
 {
@@ -400,6 +431,7 @@ QList<QueryMatch> RunnerContext::matches() const
     return matches;
 }
 
+#if KRUNNER_BUILD_DEPRECATED_SINCE(5, 79)
 QueryMatch RunnerContext::match(const QString &id) const
 {
     LOCK_FOR_READ(d)
@@ -407,6 +439,13 @@ QueryMatch RunnerContext::match(const QString &id) const
     UNLOCK(d)
     return match ? *match : QueryMatch(nullptr);
 }
+#endif
+
+void RunnerContext::requestQueryStringUpdate(const QString &text, int cursorPosition) const
+{
+    d->requestedText = text;
+    d->requestedCursorPosition = cursorPosition;
+}
 
 void RunnerContext::setSingleRunnerQueryMode(bool enabled)
 {
@@ -418,12 +457,22 @@ bool RunnerContext::singleRunnerQueryMode() const
     return d->singleRunnerQueryMode;
 }
 
+void RunnerContext::ignoreCurrentMatchForHistory() const
+{
+    d->shouldIgnoreCurrentMatchForHistory = true;
+}
+
+bool RunnerContext::shouldIgnoreCurrentMatchForHistory() const
+{
+    return d->shouldIgnoreCurrentMatchForHistory;
+}
+
 void RunnerContext::restore(const KConfigGroup &config)
 {
     const QStringList cfgList = config.readEntry("LaunchCounts", QStringList());
 
-    const QRegularExpression re(QStringLiteral("(\\d*) (.+)"));
-    for (const QString& entry : cfgList) {
+    static const QRegularExpression re(QStringLiteral("(\\d*) (.+)"));
+    for (const QString &entry : cfgList) {
         const QRegularExpressionMatch match = re.match(entry);
         if (!match.hasMatch()) {
             continue;
@@ -454,7 +503,15 @@ void RunnerContext::run(const QueryMatch &match)
     match.run(*this);
 }
 
+QString RunnerContext::requestedQueryString() const
+{
+    return d->requestedText;
+}
+int RunnerContext::requestedCursorPosition() const
+{
+    return d->requestedCursorPosition;
+}
+
 } // Plasma namespace
 
-
 #include "moc_runnercontext.cpp"
